#include "keymapwidget.h"
#include "error.h"

#include <qdragobject.h>


KeymapWidget *keymapWidget;

KeymapWidget::KeymapWidget(QWidget *parent, const char *name) :

  QWidget(parent, name), Observer(NULL) {

  setBackgroundMode(Qt::NoBackground);
  setMouseTracking(true);

  background = new QPixmap(IMAGEDIR "/keyboard.png");
  if (background->isNull()) error_message("image not found: " IMAGEDIR "/keyboard.png");

  buffer = new QPixmap(384, 48);

  barFragments[0] = new QPixmap(IMAGEDIR "/rangebar_000.png");
  if (barFragments[0]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_000.png");

  barFragments[1] = new QPixmap(IMAGEDIR "/rangebar_w00.png");
  if (barFragments[1]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_w00.png");

  barFragments[2] = new QPixmap(IMAGEDIR "/rangebar_w01.png");
  if (barFragments[2]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_w01.png");

  barFragments[3] = new QPixmap(IMAGEDIR "/rangebar_w02.png");
  if (barFragments[3]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_w02.png");

  barFragments[4] = new QPixmap(IMAGEDIR "/rangebar_w03.png");
  if (barFragments[4]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_w03.png");

  barFragments[5] = barFragments[0];

  barFragments[6] = new QPixmap(IMAGEDIR "/rangebar_r00.png");
  if (barFragments[6]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_r00.png");

  barFragments[7] = new QPixmap(IMAGEDIR "/rangebar_r01.png");
  if (barFragments[7]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_r01.png");

  barFragments[8] = new QPixmap(IMAGEDIR "/rangebar_r02.png");
  if (barFragments[8]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_r02.png");

  barFragments[9] = new QPixmap(IMAGEDIR "/rangebar_r03.png");
  if (barFragments[0]->isNull()) error_message("image not found: " IMAGEDIR "/rangebar_r03.png");

  marker = new QPixmap(IMAGEDIR "/keymarker.png");
  if (marker->isNull()) error_message("image not found: " IMAGEDIR "/keymarker.png");

  rootKeyMarker = new QPixmap(IMAGEDIR "/root_marker.png");
  if (rootKeyMarker->isNull()) error_message("image not found: " IMAGEDIR "/root_marker.png");

  memset(icons, 0, sizeof(int) * 120);

  sample = NULL;
  instrument = NULL;
  sampleMap = NULL;
  
  xOffset = 0;
  selectedKey = -1;

  setAcceptDrops(TRUE);

  resize(384, 48);
  update(NULL);
}


void KeymapWidget::subjectDeleted(Subject *s) {

  if (s == sampleMap) { 

    sampleMap = NULL;
    instrument = NULL;
    sample = NULL;

    update(NULL);
  }
}


void KeymapWidget::renderBuffer() {

  int x = 0;

  QPainter *p = new QPainter(buffer);
  p->drawPixmap(0, 0, *background, (xOffset << 3), 0, 384, 48); 

  int j = xOffset;

  for (int i = 0; i < 48; i++) {

    p->drawPixmap(x, 1, *barFragments[icons[j]], 0, 0, 8, 18);

    if (instrument) {

      Sample* s = instrument->sampleMap.getSample(j);

      if (s && (s->getMidiNote() == j)) {

	p->drawPixmap(x, 1, *rootKeyMarker, 0, 0, 8, 18);
      }
    }

    j++;
    x += 8;
  }

  int selectedPos = selectedKey - xOffset;

  if ((selectedPos >= 0) && (selectedPos < 48)) 
    p->drawPixmap(1 + selectedPos * 8, 21, *marker, 0, 0, 5, 11);

  p->end();
}


void KeymapWidget::update(Subject *s) {

  if (s != sampleMap) return;

  buildIconArray();
  renderBuffer();
  repaint();
}


void KeymapWidget::paintEvent(QPaintEvent *e) {

  QPainter *p = new QPainter(this);
  p->drawPixmap(0, 0, *buffer);
  p->end();
}


void KeymapWidget::setSample(Sample *s) {

  if (sample != s) {

    sample = s;
    update(sampleMap);
  }
}


void KeymapWidget::buildIconArray() {

  if (!sampleMap) return;

  int i = 0;
  while (i < 120) {

    Sample *s = sampleMap->getSample(i);

    if (s == NULL) {

      icons[i++] = 0;

    } else {
      
      int offset = 0;
      if (s == sample) offset = 5;
      
      int j = i + 1;
      int count = 1;
      while((j < 120) && 
	    (sampleMap->getSample(j) == s)) { count++; j++; }
      
      switch(count) {
	
      case 1:
	icons[i] = 1 + offset;
	i++;
	break;
	
      case 2:
	icons[i] = 2 + offset;
	i++;
	icons[i] = 4 + offset;
	i++;
	break;
	
      default:
	
	count -= 2;
	
	icons[i] = 2 + offset;
	i++;
	
	while(count > 0) { icons[i++] = 3 + offset; count--; }
	
	icons[i] = 4 + offset;
	i++;
      }
    }
  }
}


void KeymapWidget::mousePressEvent(QMouseEvent *e) {

  int key = xOffset + (e->x() / 8);

  if (e->y() < 18) {
    
    Sample* s;

    if (sampleMap && (s = sampleMap->getSample(key))) {

      setSample(s);
      emit(sampleChanged(s));
    }

  } else {

    switch(e->button() | e->state()) {
      
    case (Qt::LeftButton | Qt::ShiftButton):
      if (sample) sample->setMidiRangeStart(key);
      break;
      
    case (Qt::RightButton | Qt::ShiftButton):
      if (sample) sample->setMidiRangeEnd(key);
      break;
      
    case (Qt::LeftButton | Qt::ControlButton):
      if (sample) sample->setMidiNote(key);
      break;
      
    case (Qt::LeftButton):
      
      playedChannel = instrument->getMidiChannel();
      playedNote = selectedKey;
      
      Midi::noteOn(playedChannel, playedNote, 127);
      
      break;
      
    }
  }
}


void KeymapWidget::mouseReleaseEvent(QMouseEvent *e) {

  switch(e->button()) {

  case Qt::LeftButton:

    Midi::noteOff(playedChannel, playedNote);

    break;
  }

}

void KeymapWidget::wheelEvent(QWheelEvent *e) {

  if ((xOffset > 0) && (e->delta() < 0)) xOffset--;
  else if ((xOffset < 72) && (e->delta() > 0)) xOffset++;

  renderBuffer();
  repaint();
}


void KeymapWidget::mouseMoveEvent(QMouseEvent *e) {

  selectedKey = xOffset + (e->x() / 8);
  renderBuffer();
  repaint();
}


void KeymapWidget::setInstrument(Instrument *i) {

  if (sampleMap) sampleMap->removeObserver(this);
    
  if (i) {

    instrument = i;
    sampleMap = &i->sampleMap;
    sample = NULL;

    sampleMap->addObserver(this);

  } else {

    instrument = NULL;
    sampleMap = NULL;
    sample = NULL;
  }
 
  update(sampleMap);
}




void KeymapWidget::dropEvent(QDropEvent* e) {

  error_message("dropped: %s", e->format());

  QStringList stringList;

  if (QUriDrag::decodeLocalFiles(e, stringList)) {

    int i = 0;

    for (QStringList::iterator it = stringList.begin(); 
	 it != stringList.end(); it++) {

      QString fileName = *it;
      
      Sample* s = new Sample(instrument);
      
      if (s->loadFile(fileName) < 0) {
	
	error_message("could not load: %s", (const char*)fileName);
	
	delete(s);

      } else {
      
	instrument->addSample(s);
	s->setMidiNote(selectedKey + i);

	i++;
      }
    }
  }
}

void KeymapWidget::dragEnterEvent(QDragEnterEvent* e) {

  e->accept(QUriDrag::canDecode(e));
}


