#include <stdlib.h>

#include "knobwidget.h"
#include "error.h"

const float PI = 3.1415927f;

#define RAD_OFFSET (-1.375f * PI)

KnobWidget::KnobWidget(QPixmap *b, QPixmap *i, 
		       int w, int h, 
		       int ox, int oy,
		       float r,
		       QWidget *parent, const char *name) 
  : QWidget(parent, name) {

  body = b;
  indicator = i;

  iconWidth = w;
  iconHeight = h;

  radius = r;

  inOffsetX = ox;
  inOffsetY = oy;

  value = 0;

  minValue = 0;
  maxValue = 127;

  stepSize = 2;

  resize(iconWidth, iconHeight);
}

void KnobWidget::paintEvent(QPaintEvent *e) {

  calcIndicatorPos();

  QPainter *p = new QPainter(this);

  p->drawPixmap(0, 0, *body);
  p->drawPixmap(inX, inY, *indicator);

  p->end();
}


void KnobWidget::mousePressEvent(QMouseEvent *e) {
  
  x1 = (float)(e->x() - iconWidth / 2);
  y1 = (float)(e->y() - iconHeight / 2);
}

void KnobWidget::mouseReleaseEvent(QMouseEvent *e) {
}


void KnobWidget::mouseMoveEvent(QMouseEvent *e) {

  float x2 = (float)(e->x() - width() / 2);
  float y2 = (float)(e->y() - height() / 2);

  float mag = sqrtf(x2 * x2 + y2 * y2);

  x2 = x2 / mag;
  y2 = y2 / mag;
  
  float a = acos(y2);

  if (x2 > 0) a = 2.0f * PI - a;

  a -= 0.125f * PI;

  if (a < 0.0f) a = 0.0f;
  else if (a > 1.75f * PI) a = 1.75f * PI;

  float ticksPerRad = (float)abs(maxValue - minValue) / (1.75f * PI);

  value = minValue + (int)(a * ticksPerRad);
  //error_message("[%f]->[%d]", a, value);

  repaint();

  emit(valueChanged(value));
}

void KnobWidget::calcIndicatorPos() {
  
  float x = (float)(value - minValue);
  float l = (float)abs(maxValue - minValue);

  float radPerTick = (1.75f * PI) / l;

  float a = RAD_OFFSET + radPerTick * x;

  inX = inOffsetX + (int)(cos(a) * radius);
  inY = inOffsetY + (int)(sin(a) * radius);
}

void KnobWidget::wheelEvent(QWheelEvent *e) {

  if ((value + stepSize <= maxValue) && (e->delta() > 0)) value += stepSize;
  else if ((value - stepSize >= minValue) && (e->delta() < 0)) value -= stepSize;
  else return;
  
  repaint();
 
  emit(valueChanged(value));
}


void KnobWidget::setValue(int v) {

  if (v < minValue) value = minValue;
  if (v > maxValue) value = maxValue;
  else value = v;

  repaint();
}

int KnobWidget::getValue() {
  
  return(value);
}

void KnobWidget::setRange(int a, int b) {

  if (value < a) value = a;
  if (value > b) value = b;
  
  minValue = a;
  maxValue = b;

  repaint();
}

QPixmap *Knob25::body25;
QPixmap *Knob25::indicator25;

QPixmap *Knob35::body35;
QPixmap *Knob35::indicator35;

Knob25::Knob25(QWidget *parent, const char *name) :
  KnobWidget(body25, indicator25, 
	     25, 25,
	     12, 12,
	     7.0f,
	     parent,
	     name) {
}

Knob35::Knob35(QWidget *parent, const char *name) :
  KnobWidget(body35, indicator35, 
	     25, 25,
	     14, 14,
	     10.0f,
	     parent,
	     name) {
}

