#include <qwidget.h>
#include <qpainter.h>
#include <qpixmap.h>

#include <assert.h>

#include "sampleview.h"
#include "sample.h"
#include "observer.h"

#include "error.h"

#include <list>


float SampleView::W_stepSize() {

  float d = 1 + viewEnd - viewStart;
  return((float)width() / d);
}

SampleView::SampleView(QWidget *parent, const char *name) 
  : QWidget(parent, name), Observer(NULL) { 

  sample = NULL;
  pixmap = new QPixmap(384, 192);
  pixmap->fill(QColor(0x00, 0xCC, 0xFF));

  buffers[0] = (float*)malloc(sizeof(float) * 384);
  buffers[1] = (float*)malloc(sizeof(float) * 384);

  bufferLen = 384;

  if (!buffers[0] || !buffers[1]) {
    error_message("out of mem!");
    exit(110);
  }

  setBackgroundMode(Qt::NoBackground);

  selectionMode = 0;

  viewStart = 0;
  viewEnd = 0;

  scaleFactor = 1.0f;
}


void SampleView::update(Subject *s) {

  if (s != sample) return;

  repaint();
}


void SampleView::renderPixmap() {

  QPainter *p = new QPainter(pixmap);
  pixmap->fill(QColor(0x00, 0xCC, 0xFF));

  if (!sample || !sample->isValid()) return;

  int numChannels = std::min(sample->getNumChannels(), SV_NUM_BUFFERS);

  sample->renderData(buffers, numChannels, viewStart, viewEnd, bufferLen);

  p->setPen(Qt::green);

  int chHeight = height() / numChannels;
  int chOffset = chHeight >> 1;

  int y = chOffset;

  for (int j = 0; j < numChannels; j++) {
    for (int i = 0; i < bufferLen; i++) {
    
      p->drawLine(i, y, 
		  i, y + (int)(buffers[j][i] * 0.9f * scaleFactor * chOffset));
    }

    y += chHeight;
  }
  
  p->end();
}

void SampleView::subjectDeleted(Subject *s) {

  if (s == sample) {
    
    sample = NULL;
    update(NULL);
  }
}


void SampleView::setSample(Sample *s) {

  if (s == sample) { error_message("ABORT REDUNDANT CALL"); return; }

  if (sample) sample->removeObserver(this);

  sample = s;

  if (sample) { 

    sample->addObserver(this);

    viewStart = 0;
    viewEnd = std::max(0, sample->getNumFrames() - 1);

  }

  renderPixmap();
  repaint();
}


void SampleView::paintEvent(QPaintEvent *e) {

  QPainter *p = new QPainter(this);

  p->drawPixmap(0, 0, *pixmap); 

  if (!sample || !sample->isValid()) { p->end(); return; }

  int tx;

  p->setPen(Qt::red);

  tx = sampleToWindow(sample->getSampleStart());
  if (tx >= 0 && tx <= width()) p->drawLine(tx, 0, tx, height() - 1);

  tx = sampleToWindow(sample->getSampleEnd());
  if (tx >= 0 && tx <= width()) p->drawLine(tx, 0, tx, height() - 1);

  
  p->setPen(Qt::blue);

  tx = sampleToWindow(sample->getLoopStart());
  if (tx >= 0 && tx <= width()) p->drawLine(tx, 0, tx, height() - 1);

  tx = sampleToWindow(sample->getLoopEnd());
  if (tx >= 0 && tx <= width()) p->drawLine(tx, 0, tx, height() - 1);


  float sx1 = (float)(viewStart) / (float)sample->getNumFrames();
  float sx2 = (float)(viewEnd) / (float)sample->getNumFrames();

  p->setPen(Qt::black);

  p->drawLine((int)(sx1 * width()), height() / 2, 
	      (int)(sx2 * width()), height() / 2);


  p->end();
}


void SampleView::mousePressEvent(QMouseEvent *e) {

  int p = windowToSample(e->x());

  if ((p >= 0) && (p < sample->getNumFrames())) {
    
    switch(selectionMode) {
      
    case 0: // loop
      
      switch (e->button()) {

      case LeftButton:
	sample->setLoopStart(p);
	break;

      case RightButton:
	sample->setLoopEnd(p);
	break;
      }

      break;

    case 1: // trim
      
      switch (e->button()) {

      case LeftButton:
	sample->setSampleStart(p);
	break;
	
      case RightButton:
	sample->setSampleEnd(p);
	break;
      }
      
      break;
    }
  }
}  


void SampleView::setSelectionMode(int v) {

  selectionMode = v;
}


void SampleView::setLoop() {

  selectionMode = 0;
}


void SampleView::setTrim() {

  selectionMode = 1;
}





void SampleView::wheelEvent(QWheelEvent *e) {

  switch (e->state()) {

  case Qt::ShiftButton:

    if (e->delta() > 0) scaleFactor += 1.0f;
    else if (e->delta() < 0) scaleFactor -= 1.0f;

    if (scaleFactor < 1.0f) scaleFactor = 1.0f;

    break;

  case Qt::ControlButton:

    // zoom

    if (e->delta() > 0) {
      
      // zoom in
      
      int delta = viewEnd - viewStart;
      int offset = delta / 4;
      
      viewStart += offset;
      viewEnd -= offset;

    } else if (e->delta() < 0) {
      
      // zoom out
      
      int delta = viewEnd - viewStart;
      int offset = delta / 2;
      
      viewStart -= offset;
      viewEnd += offset;
      
      if (viewStart < 0) 
	viewStart = 0;
      if (viewEnd > sample->getNumFrames() - 1) 
	viewEnd = sample->getNumFrames() -1;

    }
    
    break;

  default:

    // scroll

    int offset = (int)(16.0f / (float)W_stepSize());

    if (e->delta() > 0 ) {

      if (viewEnd + offset >= sample->getNumFrames()) {

	offset = sample->getNumFrames() - 1 - viewEnd;
      }

      viewStart += offset;
      viewEnd += offset;

    } else if (e->delta() < 0 ) {

      if (viewStart - offset < 0) {

	offset = viewStart;

      }

      viewStart -= offset;
      viewEnd -= offset;
    }

    break;
  }

  renderPixmap();
  repaint();
}


void SampleView::mouseMoveEvent(QMouseEvent *e) {
}

int SampleView::windowToSample(int x) {

  int y = viewStart + (int)(x / W_stepSize());

  //  error_message("%d -> %d", x, y);
  return(y);
}

int SampleView::sampleToWindow(int x) {

  int y = (int)((x - viewStart) * W_stepSize());

  //  error_message("%d -> %d", x, y);
  return(y);
}
