#include <qfiledialog.h>
#include <qlineedit.h>
#include <qlistbox.h>
#include <qspinbox.h>
#include <qcheckbox.h>
#include <qpushbutton.h>
#include <qaction.h>
#include <qapplication.h>
#include <qslider.h>
#include <qtoolbutton.h>
#include <qcombobox.h>

#include <string.h>

#include "simsamgui.h"
#include "simsamgui.uic.h"

#include "midilearndialog.h"

#include "simsam.h"
#include "sample.h"
#include "midi.h"
#include "output.h"
#include "error.h"
#include "observer.h"
#include "patchfile.h"
#include "voice.h"
#include "settings.h"

#include "sampleview.h"
#include "adsrwidget.h"
#include "keymapwidget.h"
#include "knobwidget.h"
#include "instrumentselector.h"
#include "sampleselector.h"

#include "outputdialog.h"
#include "nameoutputdialog.uic.h"

extern QApplication* qApplication;

/* 
 *  Constructs a SimsamGui which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 */
SimsamGui::SimsamGui(QWidget* parent,  const char* name, WFlags fl )
  : SimsamGuiUI( parent, name, fl ), Observer(NULL)
{
  
  backgroundPixmap = new QPixmap(IMAGEDIR "/background.png");
  if (backgroundPixmap->isNull()) 
    error_message("image not found: " IMAGEDIR "/background.png");
  setPaletteBackgroundPixmap(*backgroundPixmap);
  
  instrument = Instrument::getInstrument(0);
  sample = instrument->getSample(0);

  sampleView->setSample(sample);
  keymapWidget->setSample(sample);
  adsrWidget->setSample(sample);
  
  sampleSelector->setInstrument(instrument);
  keymapWidget->setInstrument(instrument);

  // setup knobs ranges

  knobMaster->setRange(0, 120);
  knobMaster->setValue(100);

  knobGain->setRange(0, 400);
  knobGain->setValue(75);

  knobPan->setRange(-64, 63);
  knobPan->setValue(0);

  connect(pbFileLoad,SIGNAL(clicked()),this,SLOT(loadSample()));
  connect(leName,SIGNAL(returnPressed()),this,SLOT(setName()));
  connect(sbGain,SIGNAL(valueChanged(int)),this,SLOT(setGain(int)));
  connect(knobGain,SIGNAL(valueChanged(int)),this,SLOT(setGain(int)));
  connect(sbPan,SIGNAL(valueChanged(int)),this,SLOT(setPan(int)));
  connect(knobPan,SIGNAL(valueChanged(int)),this,SLOT(setPan(int)));
  connect(sbSampleStart,SIGNAL(valueChanged(int)),this,SLOT(setSampleStart(int)));
  connect(sbSampleEnd,SIGNAL(valueChanged(int)),this,SLOT(setSampleEnd(int)));
  connect(sbLoopStart,SIGNAL(valueChanged(int)),this,SLOT(setLoopStart(int)));
  connect(sbLoopEnd,SIGNAL(valueChanged(int)),this,SLOT(setLoopEnd(int)));
  connect(cbLoop,SIGNAL(toggled(bool)),this,SLOT(setLoop(bool)));
  connect(cbMiddleLoop,SIGNAL(toggled(bool)),this,SLOT(setMiddleLoop(bool)));
  connect(cbTrigger,SIGNAL(toggled(bool)),this,SLOT(setTrigger(bool)));
  connect(cbExclusive,SIGNAL(toggled(bool)),this,SLOT(setExclusive(bool)));
  connect(pbLearnMidi,SIGNAL(clicked()),this,SLOT(learnMidi()));
  connect(sbMidiNote,SIGNAL(valueChanged(int)), this, SLOT(setMidiNote(int)));
  connect(tbLoop, SIGNAL(clicked()), sampleView, SLOT(setLoop()));
  connect(tbTrim, SIGNAL(clicked()), sampleView, SLOT(setTrim()));
  connect(cbReverse, SIGNAL(toggled(bool)), this, SLOT(setReverse(bool)));
  connect(cbModePreset, SIGNAL(activated(int)), this, SLOT(setModePreset(int)));
  connect(sbMasterVolume,SIGNAL(valueChanged(int)),this,SLOT(setMasterVolume(int)));
  connect(knobMaster,SIGNAL(valueChanged(int)),this,SLOT(setMasterVolume(int)));
  connect(tbPlay, SIGNAL(toggled(bool)), this, SLOT(playToggled(bool)));
  connect(tbShowWave, SIGNAL(toggled(bool)), this, SLOT(showWaveToggled(bool))); 
  connect(pbPanic, SIGNAL(clicked()), this, SLOT(panicClicked()));

  connect(instrumentSelector, SIGNAL(instrumentChanged(Instrument*)),
	  this, SLOT(setInstrument(Instrument*)));
  connect(instrumentSelector, SIGNAL(instrumentChanged(Instrument*)),
	  sampleSelector, SLOT(setInstrument(Instrument*)));
  connect(instrumentSelector, SIGNAL(instrumentChanged(Instrument*)),
	  keymapWidget, SLOT(setInstrument(Instrument*)));

  connect(sampleSelector, SIGNAL(sampleSelected(Sample*)), 
	  this, SLOT(setSample(Sample*)));
  connect(sampleSelector, SIGNAL(sampleSelected(Sample*)), 
	  sampleView, SLOT(setSample(Sample*)));
  connect(sampleSelector, SIGNAL(sampleSelected(Sample*)), 
	  keymapWidget, SLOT(setSample(Sample*)));
  connect(sampleSelector, SIGNAL(sampleSelected(Sample*)), 
	  adsrWidget, SLOT(setSample(Sample*)));

  connect(keymapWidget, SIGNAL(sampleChanged(Sample*)),
	  this, SLOT(setSample(Sample*)));
  connect(keymapWidget, SIGNAL(sampleChanged(Sample*)),
	  sampleView, SLOT(setSample(Sample*)));

}


SimsamGui::~SimsamGui() {
}


void SimsamGui::subjectDeleted(Subject *s) {

  if (s == sample) {

    sample = NULL;
    update(NULL);
  }
}


void SimsamGui::panicClicked() {

  Simsam::panic();
}


void SimsamGui::setReverse(bool on) {

  if (!sample) {error_message("NULL"); return; }
  sample->setReverse(on);
}


void SimsamGui::setModePreset(int t) {

  if (!sample) {error_message("NULL"); return; }

  if (t < 2) return;

  switch(t) {

  case 2: // normal
    setLoop(false);
    setMiddleLoop(false);
    setTrigger(false);
    break;

  case 3: // loop
    setLoop(true);
    setMiddleLoop(false);
    setTrigger(false);
    break;

  case 4: // loop while pressed
    setLoop(true);
    setMiddleLoop(true);
    setTrigger(true);
    break;

  case 5: // loop until pressed
    setLoop(true);
    setMiddleLoop(false);
    setTrigger(true);
    break;

  case 6: // one shot
    setLoop(false);
    setMiddleLoop(false);
    setTrigger(true);
    break;
  }
}


void SimsamGui::update(Subject *s) {

  if (s != sample) return;

  if (s) {

    // valid update...


    int sampleMax = std::max(0, sample->getNumFrames() - 1);
    
    sbPan->setValue(sample->getPan());
    knobPan->setValue(sample->getPan());
    
    sbGain->setValue(sample->getGain());
    knobGain->setValue(sample->getGain());
    
    sbSampleStart->setMaxValue(sampleMax);
    sbSampleStart->setValue(sample->getSampleStart());
    
    sbSampleEnd->setMaxValue(sampleMax);
    sbSampleEnd->setValue(sample->getSampleEnd());
    
    sbLoopStart->setMaxValue(sampleMax);
    sbLoopStart->setValue(sample->getLoopStart());
    
    sbLoopEnd->setMaxValue(sampleMax);
    sbLoopEnd->setValue(sample->getLoopEnd());
    
    cbLoop->setChecked(sample->getLoop());
    cbMiddleLoop->setChecked(sample->getMiddleLoop());
    cbTrigger->setChecked(sample->getTrigger());
    cbExclusive->setChecked(sample->getExclusive());

    cbReverse->setChecked(sample->getReverse());
    
    sbMidiNote->setValue(sample->getMidiNote());
    
    leFilePath->setText(sample->path);
    leName->setText(sample->name);

  }



  repaint();
}


void SimsamGui::setSample(Sample* s) {

  if (sample) sample->removeObserver(this);

  sample = s;

  if (sample) { 

    sample->addObserver(this); 
    update(sample);

  } else {

    // subject was reset to NULL, set default values!

    sbPan->setValue(0);
    knobPan->setValue(0);
    
    sbGain->setValue(0);
    knobGain->setValue(0);
    
    sbSampleStart->setMaxValue(0);
    sbSampleStart->setValue(0);
    
    sbSampleEnd->setMaxValue(0);
    sbSampleEnd->setValue(0);
    
    sbLoopStart->setMaxValue(0);
    sbLoopStart->setValue(0);
    
    sbLoopEnd->setMaxValue(0);
    sbLoopEnd->setValue(0);
    
    cbLoop->setChecked(0);
    cbMiddleLoop->setChecked(0);
    cbTrigger->setChecked(0);
    cbExclusive->setChecked(0);

    cbReverse->setChecked(0);
    
    sbMidiNote->setValue(0);
    
    leFilePath->setText("");
    leName->setText("empty!");
  }

}

void SimsamGui::setInstrument(Instrument* i) {

  instrument = i;
  setSample(instrument->getSample(0));
}

void SimsamGui::loadSample() {

  if (!sample) {error_message("NULL"); return; }

  QString s (QFileDialog::getOpenFileName(Settings::samplePath,
					  "Sample files (*.wav *.aiff *.aifc *.snd *.au *.voc)", 
					  this ));
  
  if (s.isEmpty()) return;

  if (sample->loadFile(s) < 0) {

    exit(10);
  }
  
  strncpy(sample->path,s,1023);

  sampleView->setSample(sample);
}


void SimsamGui::setGain(int gain) {
  
  if (!sample) {error_message("NULL"); return; }

  sample->setGain(gain);
}


void SimsamGui::setName() {
  
  if (!sample) {error_message("NULL"); return; }

  QString s = leName->text();
  strncpy(sample->name,s,255);
}


void SimsamGui::setPan(int val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setPan(val);
}


void SimsamGui::setSampleStart(int val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setSampleStart(val);
}


void SimsamGui::setSampleEnd(int val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setSampleEnd(val);
}


void SimsamGui::setLoopStart(int val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setLoopStart(val);
}


void SimsamGui::setLoopEnd(int val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setLoopEnd(val);
}


void SimsamGui::setLoop(bool val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setLoop(val);
}


void SimsamGui::setMiddleLoop(bool val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setMiddleLoop(val);
}


void SimsamGui::setTrigger(bool val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setTrigger(val);
}


void SimsamGui::setExclusive(bool val) {

  if (!sample) {error_message("NULL"); return; }

  sample->setExclusive(val);
}


void SimsamGui::learnMidi() {

  if (!sample) {error_message("NULL"); return; }

  MidiLearnDialog *d = new MidiLearnDialog();

  d->setNote(sample->getMidiNote());
  d->show();

  Midi::lastNotePlayed = -1;

  while(d->running){

    if (Midi::lastNotePlayed != -1) d->setNote(Midi::lastNotePlayed);

    qApplication->processEvents();
  }


  if (d->result() == QDialog::Accepted) {

    int i = d->getNote();
    sample->setMidiNote(i);
  }

  delete(d);
}


void SimsamGui::setMidiNote(int v) {

  if (!sample) {error_message("NULL"); return; }

  if (sample->getMidiNote() != v) sample->setMidiNote(v);
}


void SimsamGui::setMasterVolume(int val) {

  Output::masterVolume = (float)val / 100.0f;
  Output::masterVolume *=  Output::masterVolume;

  if (sbMasterVolume->value() != val) sbMasterVolume->setValue(val);
  if (knobMaster->getValue() != val) knobMaster->setValue(val);
}


void SimsamGui::playToggled(bool on) {

  if (!sample) {error_message("NULL"); return; }

  playedNote = sample->getMidiNote();
  playedChannel = instrument->getMidiChannel();
  
  if (on) {
    
    Midi::noteOn(playedChannel, playedNote, 127);
    
  } else { 
    
    Midi::noteOff(playedChannel, playedNote);
    
  }
}


void SimsamGui::showWaveToggled(bool on) {

  if (on) sampleView->show();
  else sampleView->hide();
}




SimsamMainWindow::SimsamMainWindow() 
  : QMainWindow() {
  
  setCaption("simsam!");
  
  QPopupMenu *fileMenu = new QPopupMenu(this);
  menuBar()->insertItem("&File", fileMenu);

  fileMenu->insertItem( "&Open", this, SLOT(load()), CTRL+Key_O );
  fileMenu->insertItem( "&Save", this, SLOT(save()), CTRL+Key_N );
  fileMenu->insertSeparator();
  fileMenu->insertItem( "E&xit", qApplication, SLOT(quit()), CTRL+Key_X );

  QPopupMenu *configMenu = new QPopupMenu(this);
  menuBar()->insertItem("&Config", configMenu);

  configMenu->insertItem("Outputs", this, SLOT(outputConfig()));
  // configMenu->insertItem("GUI", this, SLOT(guiConfig()));

  gui = new SimsamGui(this, "simsamgui");
  gui->setFocus();

  setCentralWidget(gui);  

  resize(760, 600);
}
 
SimsamMainWindow::~SimsamMainWindow() {
}

void SimsamMainWindow::load() {

  int i;

  QString s (QFileDialog::getOpenFileName (Settings::patchPath, 
					   "*.xml", this ));
  if (s.isEmpty()) return;

  if (Simsam::load(s) < 0) {
    
    error_message("error: loading failed!");
    exit(20);
  }

}


void SimsamMainWindow::outputConfig() {
  
  OutputDialog *d = new OutputDialog();
  d->exec();
}

/*
void SimsamMainWindow::guiConfig() {
  
  GuiDialog *d = new GuiDialog();
  d->exec();
}
*/

void SimsamMainWindow::save() {

  QString s (QFileDialog::getSaveFileName (Settings::patchPath,
					   "*.xml", this ));
  if (s.isEmpty()) return;

  if (Simsam::save(s) < 0) {
    
    error_message("error: saving failed!");
    exit(30);
  }
}



