#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "envelope.h"
#include "error.h"
#include "output.h"
#include "simsam.h"

EnvADSR::EnvADSR() {

  state = OFF;
  value = 0.0f;
  delta = 0.0f;

  attackLevel = 1.0f;
  iAttackLevel = 100;

  sustainLevel = 1.0f;
  iSustainLevel = 100;

  attackTimeMS = 0;
  decayTimeMS = 50;
  releaseTimeMS = 100;
  
  Simsam::msToFrames(attackTimeMS, &attackTime);
  Simsam::msToFrames(decayTimeMS, &decayTime);
  Simsam::msToFrames(releaseTimeMS, &releaseTime);
}


void EnvADSR::noteOn() {

  value = 0.0f;
  counter = 0;
  state = ATTACK;
  calcDeltas();
}

void EnvADSR::noteOff() {

  value = sustainLevel;
  counter = 0;
  state = RELEASE;
  calcDeltas();
}

bool EnvADSR::tick(float *s) {

  switch(state) {

  case OFF:
    *s = 0.0f;
    return(false);

  case ATTACK:
    
    if (counter >= attackTime) {

      value = attackLevel;
      counter = 0;
      state = DECAY;
      calcDeltas();
      break;
    }

    value += delta;
    counter++;
    break;

  case DECAY:

    if (counter >= decayTime) {

      value = sustainLevel;
      counter = 0;
      state = SUSTAIN;
      calcDeltas();
      break;
    }

    value += delta;
    counter++;
    break;
    
  case SUSTAIN:

    break;

  case RELEASE:
    
    if (counter >= releaseTime) {

      value = 0.0f;
      state = OFF;
      break;
    }

    value += delta;
    counter++;
    break;

  default:
    *s = 0.0f;
    return(false);
  }

  *s = value;

  return(true);
}


void EnvADSR::calcDeltas() {

  switch(state) {

  case OFF:
    delta = 0.0f;
    break;

  case ATTACK:
    delta = (float)(attackLevel - value) / (float)(attackTime);
    break;

  case DECAY:
    delta = (float)(sustainLevel - value) / (float)(decayTime);
    break;

  case SUSTAIN:
    delta = 0.0f;
    break;

  case RELEASE:
    delta = (float)(0.0f - value) / (float)(releaseTime);
    break;
  }
}


void EnvADSR::resetCounter() {

  state = ATTACK;
  value = 0.0f;
  counter = 0;

  calcDeltas();
}


void EnvADSR::setAttackTime(int v) {

  attackTimeMS = v;
  Simsam::msToFrames(v, &attackTime);

  if (attackTime < 1) attackTime = 1;
  
  calcDeltas();
  notify();
}


void EnvADSR::setAttackLevel(int v) {

  iAttackLevel = v;
  attackLevel = (float)v / 100.0f;
  
  calcDeltas();
  notify();
}

void EnvADSR::setDecayTime(int v) {

  decayTimeMS = v;
  Simsam::msToFrames(v, &decayTime);

  if (decayTime < 1) decayTime = 1;
  
  calcDeltas();
  notify();
}

void EnvADSR::setSustainLevel(int v) {
  
  iSustainLevel = v;
  sustainLevel = (float)v / 100.0f;
  
  calcDeltas();
  notify();
}

void EnvADSR::setReleaseTime(int v) {

  releaseTimeMS = v;
  Simsam::msToFrames(v, &releaseTime);

  if (releaseTime < 1) releaseTime = 1;
  
  calcDeltas();
  notify();
}



int EnvADSR::getAttackTime() {

  return(attackTimeMS);
}
int EnvADSR::getDecayTime() {

  return(decayTimeMS);
}

int EnvADSR::getReleaseTime() {

  return(releaseTimeMS);
}

int EnvADSR::getAttackLevel() {

  return(iAttackLevel);
}

int EnvADSR::getSustainLevel() {

  return(iSustainLevel);
}


void EnvADSR::copyTo(EnvADSR *to) {

  memcpy(to, this, sizeof(EnvADSR));
}



/*
  $Id: envelope.cpp,v 1.3 2004-02-01 05:17:37 chenz Exp $
*/
