#include <stdlib.h>
#include <stdio.h>

#define ALSA_PCM_OLD_HW_PARAMS_API
#define ALSA_PCM_OLD_SW_PARAMS_API
#include <alsa/asoundlib.h>

#include "midi.h"
#include "simsam.h"
#include "sample.h"
#include "error.h"
#include "settings.h"
#include "output.h"

int Midi::lastNotePlayed = -1;

InstrumentMap Midi::instrumentMap;

snd_seq_t* Midi::seqHandle;
pthread_t Midi::threadHandle; 


int Midi::init () {

  /* init alsa client */

  if (snd_seq_open(&seqHandle, "default", SND_SEQ_OPEN_INPUT, 0) < 0) {
    error_message("could not open sequencer!");
    return(-1);
  }

  snd_seq_set_client_name(seqHandle, Settings::alsaSeqClientName);

  if (snd_seq_create_simple_port(seqHandle,Settings::alsaSeqPortName,
	SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
        SND_SEQ_PORT_TYPE_APPLICATION) < 0) {
    error_message("could not create sequencer port!");
    return(-1);
  }

  /* start midi thread */

  int res = pthread_create(&threadHandle, NULL, midiThread, NULL);
  if (res != 0) { 
    error_message("could not create midi-thread!");
    return(-1);
  }

  return 0;
}


int Midi::update () {

  snd_seq_event_t *ev;

  int note;
  int velocity;
  int channel;

  do {
    
    snd_seq_event_input(seqHandle, &ev);
    
    switch (ev->type) {

    case SND_SEQ_EVENT_NOTEON:

      note = ev->data.note.note;
      velocity = ev->data.note.velocity;
      channel = ev->data.note.channel;

      lastNotePlayed = note;

      noteOn(channel, note, velocity);

      break;        
      
    case SND_SEQ_EVENT_NOTEOFF:

      note = ev->data.note.note;
      channel = ev->data.note.channel;

      noteOff(channel, note);
      
      break;
    }
    
    snd_seq_free_event(ev);
    
  } while (snd_seq_event_input_pending(seqHandle, 0) > 0);
  
  return (0);
}

void Midi::noteOn(int channel, int note, int velocity) {

  error_message("%d %d %d", channel, note, velocity);

  Instrument* i = NULL;
  Sample* s = NULL;
  Voice* v = NULL;

  v = Voice::getVoice(channel, note);
	
  if (velocity > 0) {	
	
    if (v) {
      
      // re-trigger existing voice
      
      v->noteOn(channel, note, velocity);
      
    } else {
      
      // voice == NULL, start new voice
      
      i = instrumentMap.getInstrument(channel);
      if (!i) return; // no instrument mapped on channel
      
      s = i->sampleMap.getSample(note);
      if (!s) return; // no sample mapped on note 
      
      v = Voice::allocateVoice(s, i->getOutputNum());
      if (!v) return; // voice overflow
      
      v->noteOn(channel, note, velocity);
    }
    
  } else {
    
    // velocity == 0
    if (v) v->noteOff();
  }
}


void Midi::noteOff(int channel, int note) {

  error_message("%d %d %d", channel, note);
  
  Voice* v = Voice::getVoice(channel, note);
  if (v) v->noteOff();
}


void* Midi::midiThread(void *v) {

  int seq_nfds, i;
  struct pollfd *pfds;
 
  seq_nfds = snd_seq_poll_descriptors_count(seqHandle, POLLIN);

  pfds = (struct pollfd *) alloca (sizeof(struct pollfd) * (seq_nfds));

  snd_seq_poll_descriptors(seqHandle, pfds, seq_nfds, POLLIN);

  while (1) {

    if (poll(pfds, seq_nfds, 1000) > 0) {

      for (i = 0; i < seq_nfds; i++) {

	if (pfds [i].revents > 0) update();
      }
    }
  }
}






















InstrumentMap::InstrumentMap() {
  
  clearAll();
}

void InstrumentMap::setInstrument(int channel, Instrument *s) {

  if (instruments[channel] != NULL) return; 

  clearInstrument(s);
  instruments[channel] = s;
}

Instrument* InstrumentMap::getInstrument(int channel) {
  
  return(instruments[channel]);
}

void InstrumentMap::clearInstrument(Instrument *s) {
  
  for (int i = 0; i < 16; i++) if (instruments[i] == s) instruments[i] = NULL;
}

void InstrumentMap::clearAll() {
  
  memset(instruments, 0, sizeof(Instrument*) * 16);
}






































