#define ALSA_PCM_OLD_HW_PARAMS_API
#define ALSA_PCM_OLD_SW_PARAMS_API
#include <alsa/asoundlib.h>

#include <jack/jack.h>

#include "simsam.h"
#include "output.h"

#include "sample.h"
#include "settings.h"

#include "error.h"

#include <list>
#include <string>

#include <qdom.h>
#include <qfile.h>

#include <algorithm>

/* generic stuff */

float Output::masterVolume = 1.0f;
int Output::sampleRate = 44100;
int Output::bufferLen;

OutputMap Output::outputMap;
std::list<Output*> Output::outputList;

int Output::numOutputs = 0;

Subject Output::outputSubject;

int Output::type = -1;

Output* Output::getOutput(int outNum) {

  if (outNum >= outputList.size()) return(NULL);

  std::list<Output*>::iterator i = outputList.begin();

  while (outNum > 0) { i++; outNum--; }
  return(*i);
}


Output* Output::getDefaultOutput() {
  
  Output* o = getOutput(0);
  if (!o) { error_message("no default output available!"); exit(1); }
  
  return(o);
}

void Output::removeOutput(int outNum) {

  Output* o = getOutput(outNum);
  if (o) removeOutput(o);
}


int Output::addOutput(const char* n) {

  switch(type) {

  case OUTPUT_ALSA:
    return(OutputAlsa::addOutput(n));
    
  case OUTPUT_JACK:
    return(OutputJack::addOutput(n));
  }

  return(-1);
}

void Output::removeOutput(Output* o) {

  if (o == outputList.front()) {
    
    error_message("cannot remove default output!!!");
    return;
  }

  outputList.remove(*find(outputList.begin(), outputList.end(), o));
  if (numOutputs > 1) numOutputs--;
  delete o;

  outputSubject.notify();
}

int Output::init(int t) {

  int res;

  switch(t) {
    
  case OUTPUT_ALSA:
    
    res = OutputAlsa::subInit();
    break;
    
  case OUTPUT_JACK:

    res = OutputJack::subInit();
    break;
  }

  type = t;


  // try to create additional outputs as listed in ~/.simsam/outputs.xml

  char buffer[1024];
  
  memset(buffer, 0, 1024);

  char* homeDir = getenv("HOME");
  
  snprintf(buffer, 1023, "%s/.simsam/outputs.xml", homeDir);

  // load settings.xml

  QDomDocument doc("outputs");
  QFile file(buffer);

  if (!file.open(IO_ReadOnly)) return(res);

  if (!doc.setContent(&file)) {
    file.close();
    failure("doc.setContent");
  }

  file.close();

  QDomElement docElem = doc.documentElement();
  QDomNodeList sList = docElem.childNodes();

  for (int i = 0; i < sList.length(); i++) {
  
    QDomNode node = sList.item(i);
    QDomElement element = node.toElement();

    if (element.tagName() == "output") {

      addOutput(strdup(element.text()));
    }
  }


  return(res);
}


int Output::getNumOutputs() {

  return(numOutputs);
}


/* generic object methods */

Output::~Output() {
}

const char* Output::getName() {

  return(name);
}

void Output::setName(const char* nName) {

  name = nName;
}


/* ALSA module */

pthread_t OutputAlsa::threadHandle;

snd_pcm_t* OutputAlsa::pcmHandle;
snd_pcm_hw_params_t* OutputAlsa::pcmParams;

float* OutputAlsa::buffer_l;
float* OutputAlsa::buffer_r;

short* OutputAlsa::outputBuffer;

OutputAlsa::OutputAlsa(const char* nName) {

  name = nName;
}

OutputAlsa::~OutputAlsa() {
}

int OutputAlsa::subInit () { 

  bufferLen = Settings::alsaBufferLength;

  outputBuffer = (short*)malloc(sizeof(short)*  2*  bufferLen);

  buffer_l = (float*)malloc(sizeof(float)*  bufferLen);
  buffer_r = (float*)malloc(sizeof(float)*  bufferLen);

  if (!(outputBuffer && buffer_l && buffer_r)) { error_message("buffer allocation"); return(-1); }

  int res;

  /* open alsa pcm */

  if (snd_pcm_open(&pcmHandle, Settings::alsaPcmDevice, SND_PCM_STREAM_PLAYBACK, 0) < 0) {
    error_message("could not open pcm device '%s'!" ,Settings::alsaPcmDevice);
    return(-1);
  }

  /* set params */

  snd_pcm_hw_params_t* hw_params;
  snd_pcm_sw_params_t* sw_params;

  snd_pcm_hw_params_alloca(&hw_params);
  snd_pcm_hw_params_any(pcmHandle, hw_params);
  snd_pcm_hw_params_set_access(pcmHandle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED);
  snd_pcm_hw_params_set_format(pcmHandle, hw_params, SND_PCM_FORMAT_S16_LE);
  snd_pcm_hw_params_set_rate_near(pcmHandle, hw_params, Settings::alsaSampleRate, 0);
  snd_pcm_hw_params_set_channels(pcmHandle, hw_params, 2);
  snd_pcm_hw_params_set_periods(pcmHandle, hw_params, 2, 0);
  snd_pcm_hw_params_set_period_size(pcmHandle, hw_params, Settings::alsaBufferLength, 0);
  snd_pcm_hw_params(pcmHandle, hw_params);

  snd_pcm_sw_params_alloca(&sw_params);
  snd_pcm_sw_params_current(pcmHandle, sw_params);
  snd_pcm_sw_params_set_avail_min(pcmHandle, sw_params, Settings::alsaBufferLength);
  snd_pcm_sw_params(pcmHandle, sw_params);

  /* create audio thread */
  /* FIXME: run in realtime if possible */
  res = pthread_create (&threadHandle, NULL, OutputAlsa::audioThread, NULL);
  if (res != 0) {
    
    error_message ("could not create audio-thread!");
    return(res);
  }

  Output *output = new OutputAlsa("main");

  outputList.push_front(output);
  numOutputs = 1;

  return(0);
}


inline int OutputAlsa::write (int frames) {

  int err;
  int i;

  memset(buffer_l, 0, sizeof(float) * bufferLen); 
  memset(buffer_r, 0, sizeof(float) * bufferLen); 
  
  for  (i = 0; i < SIMSAM_MAX_VOICES; i++) {
    
    Voice::voices[i].addToMix(buffer_l, buffer_r, bufferLen);
  }
  
  int j = 0;
  for (i = 0; i < bufferLen; i++) {
    
    outputBuffer[j++] = (short)(buffer_l[i]*  SAMPLE_DIVIDER*  masterVolume);
    outputBuffer[j++] = (short)(buffer_r[i]*  SAMPLE_DIVIDER*  masterVolume);
  }
  
  err = snd_pcm_writei(pcmHandle, outputBuffer, bufferLen);

  return (err);
}


void* OutputAlsa::audioThread (void* vp) {

  int nfds, i;
  struct pollfd* pfds;

  nfds = snd_pcm_poll_descriptors_count (pcmHandle);
  pfds = (struct pollfd* ) alloca (sizeof (struct pollfd)*  (nfds));

  snd_pcm_poll_descriptors (pcmHandle, pfds, nfds);

  while (1) {

    if (poll (pfds, nfds, 1000) > 0) {

      for (i = 0; i < nfds; i++) {  
  
	if (pfds [i].revents > 0) { 
	
	  if (write (bufferLen) < bufferLen) {
	  
	    fprintf (stderr, "xrun !\n");
	    snd_pcm_prepare(pcmHandle);
	  }
	}
      }        
    }
  }
}


int OutputAlsa::addOutput(const char* newName) {

  outputSubject.notify();
  return(ERR_MAX_OUTPUTS);
}

int OutputAlsa::getMaxOutputs() {

  return(1);
}

/* JACK module */

OutputJack::OutputJack(const char* newName) {

  char buffer[128];

  snprintf(buffer, 128, "%s-L", newName);
  portLeft  = jack_port_register(client, buffer, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
  
  if(!portLeft) throw(1);
 
  snprintf(buffer, 128, "%s-R", newName);
  portRight = jack_port_register(client, buffer, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);

  if(!portRight) throw(1);
  
  name = newName;
}


OutputJack::~OutputJack() {

  jack_port_unregister(client, portLeft);
  jack_port_unregister(client, portRight);
}

/* static stuff */

jack_client_t* OutputJack::client = NULL;

void OutputJack::processOutput(int nFrames) {

  jack_default_audio_sample_t* l = 
    (jack_default_audio_sample_t*)jack_port_get_buffer(portLeft, nFrames);
  jack_default_audio_sample_t* r = 
    (jack_default_audio_sample_t*)jack_port_get_buffer(portRight, nFrames);
  
  memset(l, 0, sizeof(jack_default_audio_sample_t)*  nFrames);
  memset(r, 0, sizeof(jack_default_audio_sample_t)*  nFrames);

  for (int i = 0; i < SIMSAM_MAX_VOICES; i++) {
 
    Voice* v = &Voice::voices[i];
   
    if (v->isUsed() && this == v->getOutput())
      v->addToMix(l, r, nFrames);
  } 

  for (int i = 0; i < nFrames; i++) {

    l[i] *= masterVolume;
    r[i] *= masterVolume;
  }
}


int OutputJack::processCallback(jack_nframes_t nFrames, void* v) { 

  std::list<Output*>::iterator i;

  for (i = outputList.begin(); i != outputList.end(); i++) {

    ((OutputJack*)*i)->processOutput(nFrames);
  }

  return(0);
}


int OutputJack::samplerateCallback(jack_nframes_t a, void* b) { 

  Output::sampleRate = a;
  return(0); 
}


int OutputJack::buffersizeCallback(jack_nframes_t a, void* b) { 

  bufferLen = a;
  return(0); 
}


void OutputJack::shutdownCallback(void* b) {

  error_message("JACK shutdown!");
}


int OutputJack::subInit() {

  client = jack_client_new("simsam");
  if (!client) { 
    error_message("error: could not create jack client!");    
    return(-1);
  }         
  
  jack_set_process_callback(client, processCallback, 0);
  jack_set_sample_rate_callback(client, samplerateCallback, 0);
  jack_set_buffer_size_callback(client, buffersizeCallback, 0);
  jack_on_shutdown(client, shutdownCallback, 0);
  
  if (jack_activate(client)) {

    error_message("error: could not activate jack client!");
    return(-1);
  }

  // create default output 'main'
  OutputJack* o = new OutputJack("main");
  outputList.push_front(o);
  numOutputs = 1;

  if (0 != jack_connect(client, jack_port_name(o->portLeft), Settings::jackLeftConnection))
    error_message("could not connect to port %s!", Settings::jackLeftConnection);

  if (0 != jack_connect(client, jack_port_name(o->portRight), Settings::jackRightConnection))
    error_message("could not connect to port %s!", Settings::jackRightConnection);
  return(0);
}


int OutputJack::addOutput(const char* newName) {

  if (numOutputs >= 16) return(-1);


  OutputJack* o;

  try {

    o = new OutputJack(newName);

  } catch(int e) {

    return(ERR_OTHER);
  }

  outputList.push_back(o);

  numOutputs++;
  
  outputSubject.notify();
  return(0);
}

int OutputJack::getMaxOutputs() {

  return(16);
}



/* output map */

OutputMap::OutputMap() {

  clearAll();
}

void OutputMap::setOutput(int instrument, Output* o) {

  outputs[instrument] = o;
}

void OutputMap::setOutput(int instrument, int outNum) {
  
  outputs[instrument] = Output::getOutput(outNum);
}

Output* OutputMap::getOutput(int instrument) {

  return(outputs[instrument]);
}

void OutputMap::clearOutput(Output* o) {

  for (int i = 0; i < 16; i++) if (outputs[i] == o) outputs[i] = NULL;
}

void OutputMap::clearAll() {

  memset(outputs, 0, sizeof(Output*) * 16);
}




