#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <math.h>

#include <set>

#include <sndfile.h>
#include <samplerate.h>

#include <jack/jack.h>

#include "simsam.h"
#include "sample.h"
#include "error.h"
#include "midi.h"
#include "output.h"

int Sample::loadFile(const char *filename) {

  SNDFILE *sf;
  SF_INFO sfi;

  int res;

  memset (&sfi, 0, sizeof(SF_INFO));

  sf = sf_open(filename, SFM_READ, &sfi);
  if (!sf) {
    error_message("sf_open: %s", sf_strerror(NULL));
    return(-1);
  }
  
  if (sfi.channels > 2) {

    error_message("sample has too much (>2) channels!");
    return(-1);
  }

  float *temp = (float*)malloc(sizeof(float) * sfi.channels * sfi.frames);
  if(!temp) {
    perror("malloc");
    error_message("out of memory!");
    return(-1);
  }
  
  res = sf_readf_float(sf, temp, sfi.frames);
  if(res < sfi.frames) {
    fprintf(stderr, "%s", sf_strerror(sf));
    error_message("read error!");
    return(-1);
  }

  sf_close(sf);

  int tmpNumFrames = sfi.frames;

  if (sfi.samplerate != Output::sampleRate) {
    
    // resampling to output samplerate...

    SRC_DATA rsd;

    rsd.src_ratio = (double)Output::sampleRate / (double)sfi.samplerate;

    rsd.input_frames = sfi.frames;
    rsd.output_frames = (long)(sfi.frames * rsd.src_ratio);
    
    float *resampled = (float*)malloc(sizeof(float) * sfi.channels * rsd.output_frames);
    if(!resampled) {
      perror("malloc");
      error_message("malloc");
      return(-1);
    }
    
    rsd.data_in = temp;
    rsd.data_out = resampled;

    res = src_simple(&rsd, SRC_SINC_BEST_QUALITY, sfi.channels);
    if(res != 0) {
      fprintf(stderr, "%s", src_strerror(res));
      error_message("src_simple");

      free(resampled);
      free(temp);

      return(-1);
    }

    tmpNumFrames = rsd.output_frames;

    free(temp);
    temp = resampled;
  }

  switch(sfi.channels) {

  case 1:

    lock();
    //Simsam::lock();

    if(isValid()) freeBuffers();    

    W_data[0] = temp;
    W_data[1] = NULL;

    numChannels = sfi.channels;
    numFrames = tmpNumFrames;

    unlock();
    //Simsam::unlock();

    break;

  case 2:

    float *tempL = (float*)malloc(sizeof(float) * tmpNumFrames);
    float *tempR = (float*)malloc(sizeof(float) * tmpNumFrames);

    if (!(tempL && tempR)) {

      perror("malloc");
      error_message("out of mem");

      free(temp);
      return(-1);
    }
    
    // de-interleave

    int j = 0;
    for (int i = 0; i < tmpNumFrames; i++) {
      
      tempL[i] = temp[j++];
      tempR[i] = temp[j++];
    }
    
    free(temp);

    lock();

    if(isValid()) freeBuffers();    

    W_data[0] = tempL;
    W_data[1] = tempR;

    numChannels = sfi.channels;
    numFrames = tmpNumFrames;
    
    unlock();

    break;
  }

  strncpy(path, filename, 1024);

  setPan(0);
  setGain(75);
  setSampleStart(0);
  setSampleEnd(numFrames - 1);
  setLoopStart(0);
  setLoopEnd(numFrames - 1);

  return(0);
}


void Sample::freeBuffers() {

  if(W_data[0]) free(W_data[0]);
  if(W_data[1]) free(W_data[1]);

  W_data[0] = W_data[1] = NULL;
  
  numChannels = 0;
  numFrames = 0;
}


void Sample::unload () {

  lock();

  if (W_data[0]) {

    freeBuffers();
  }

  setGain(0);
  setPan(0);
  setSampleStart(0);
  setSampleEnd(0);
  setLoopStart(0);
  setLoopEnd(0);
  setLoop(false);
  setMiddleLoop(false);
  setTrigger(false);
  setExclusive(false);
  setReverse(false);

  strcpy(name, "empty!");
  strcpy(path, "empty!");

  unlock();
}


Sample::Sample(Instrument *i) {

  id = Sample::getNewId();

  W_data[0] = NULL;
  W_data[1] = NULL;

  loop = middleLoop = trigger = exclusive = reverse = false;

  gain = 0.0f;
  iGain = 0;

  pan = 0.5f;
  iGain = 50;

  sampleStart = sampleEnd = loopStart = loopEnd = 0;

  numFrames = numChannels = 0;

  midiNote = midiRangeStart = midiRangeEnd = -1;

  strcpy (path, "empty!");
  strcpy (name, "empty!");

  pthread_mutex_init(&mutex, NULL);

  instrument = i;
}


Sample::~Sample () {

  unload ();
}


inline bool Sample::advanceReadIndex(int &i, int state) {

  if (!reverse) {
    
    i++;
    
    if (loop) {
      
      if (i > loopEnd) {
	
	if (middleLoop && (state == Voice::RELEASED)) {}
	else {
	  
	  i = loopStart;
	}
      }
    }
    
    if (i > numFrames - 1) {
      
      i = sampleStart;
      return(false);
    }
    
  } else {

    // reverse !
    
    i--;
    
    if (loop) {
      
      if (i < loopStart) {
	
	if (middleLoop && (state == Voice::RELEASED)) {}
	else {
	  
	  i = loopEnd;
	}
      }
    }
    
    if (i < 0) {
      
      i = sampleEnd;
      return(false);
    }
  }

  return(true);
}

inline bool Sample::advanceReadIndex(float &fIndex, float stepSize, int state) {

  if (!reverse) {
    
    fIndex += stepSize;
    
    if (loop) {
      
      if (fIndex > (float)loopEnd) {
	
	if (middleLoop && (state == Voice::RELEASED)) {}
	else {
	  
	  fIndex = (float)loopStart;
	}
      }
    }
    
    if (fIndex > (float)(numFrames - 1)) {
      
      fIndex = (float)sampleStart;
      return(false);
    }
    
  } else {

    // reverse !
    
    fIndex -= stepSize;
    
    if (loop) {
      
      if (fIndex < (float)loopStart) {
	
	if (middleLoop && (state == Voice::RELEASED)) {}
	else {
	  
	  fIndex = (float)loopEnd;
	}
      }
    }
    
    if (fIndex < (float)0) {
      
      fIndex = (float)sampleEnd;
      return(false);
    }
  }

  return(true);
}



inline float Sample::getMonoFrame(int i) {

  return(W_data[0][i] * gain);
}

inline void Sample::getStereoFrame(float *l, float *r, int i) {

  *l = W_data[0][i] * gain;
  *r = W_data[1][i] * gain;
}



void Sample::renderData(float **buffers, int numBuffers, int start, int end, int bufferLen) {

  if (!isValid()) return;

  lock();

  numBuffers = std::min(numBuffers, numChannels);
  
  if (bufferLen <= 0) { error_message("bufferLen <= 0"); unlock(); return; }
  
  int windowLen = end - start;
  float stepSize = (float)windowLen / (float)bufferLen;

  for (int j = 0; j < numBuffers; j++) {

    memset(buffers[j], 0, sizeof(float) * bufferLen);

    for (int i = 0; i < bufferLen; i++) {

      int k = start + (int)(i * stepSize);
      if (k >= numFrames) k = numFrames - 1;

      buffers[j][i] = W_data[j][k];
    }
  }

  unlock();
}


void Sample::setSampleStart(int v) {
  
  if ((v != sampleStart) &&
      (v >= 0) && 
      (v < numFrames)) {
    
    sampleStart = v;
    notify();
  }
}


void Sample::setSampleEnd(int v) {
  
  if ((v != sampleEnd) &&
      (v > sampleStart) && 
      (v < numFrames)) {
    
    sampleEnd = v;
    notify();
  }
}

void Sample::setLoopStart(int v) {
  
  if ((v != loopStart) &&
      (v >= 0) && 
      (v < loopEnd)) {
    
    loopStart = v;
    notify();
  }
}

void Sample::setLoopEnd(int v) {
  
  if ((v != loopEnd) &&
      (v > loopStart) && 
      (v < numFrames)) {
    
    loopEnd = v;
    notify();
  }
}



int Sample::getNumChannels() {
  
  //    error_message("NC %d", num_channels);
  return(numChannels);
}


int Sample::getNumFrames() {
  
  //    error_message("NF %d", numFrames);
  return(numFrames);
}



bool Sample::isValid() { 
  
  if ((numFrames > 0) &&
      (numChannels > 0) &&
      (W_data[0] != NULL))  return(true);
  
  return false;
}


int Sample::getLoopStart() {

  return(loopStart);
}

int Sample::getLoopEnd() {

  return(loopEnd);
}


int Sample::getSampleStart() {

  return(sampleStart);
}

int Sample::getSampleEnd() {

  return(sampleEnd);
}



void Sample::setPan(int v) {

  iPan = v;
  pan = (float)(v + 64) / 127.0f; 

  notify();
}

int Sample::getPan() {

  return(iPan);
}

void Sample::setGain(int v) {

  iGain = v;
  gain = (float)v / 100.0f;
  notify();
}

int Sample::getGain() {

  return(iGain);
}
  
void Sample::setLoop(bool on) {

  loop = on;
  notify();
}

void Sample::setMiddleLoop(bool on) {

  middleLoop = on;
  notify();
}

void Sample::setTrigger(bool on) {

  trigger = on;
  notify();
}

void Sample::setExclusive(bool on) {

  exclusive = on;
  notify();
}


bool Sample::getLoop() {

  return(loop);
}

bool Sample::getMiddleLoop() {

  return(middleLoop);
}

bool Sample::getTrigger() {

  return(trigger);
}

bool Sample::getExclusive() {

  return(exclusive);
}


void Sample::setReverse(bool on) {

  reverse = on;
  notify();
}

bool Sample::getReverse() {

  return reverse;
}


int Sample::maxId = 0;
int Sample::getNewId() {

  return(++maxId);
}

void Sample::setMidiNote(int note) {

  instrument->sampleMap.clearSample(this);
  instrument->sampleMap.setSample(note, this);
  
  midiNote = midiRangeStart = midiRangeEnd = note;

  notify();
}

int Sample::getMidiNote() {

  return(midiNote);
}


inline void Sample::lock() {
  
  pthread_mutex_lock(&mutex);
}

inline void Sample::unlock() {
  
  pthread_mutex_unlock(&mutex);
}


bool Sample::addToMix(float *l, float *r, int nFrames, int &rIndex, int state, 
		      float velocity, EnvADSR *vAdsr, int noteOffset) {

  int i;
  float sample;
  bool res = true;

  float s;

  float fIndex = (float)rIndex;
  float stepSize = powf(2.0f, (float)noteOffset / 12.0f);

  if (!isValid()) return(false);

  // protect buffer
  lock();

  switch(numChannels) {
    
  case 1:
    
    for (i = 0; i < nFrames; i++) {

      if (!(res = vAdsr->tick(&s))) break;
      
      sample = getMonoFrame((int)fIndex) * velocity * s;

      l[i] += sample * sqrt(1.0f - pan);
      r[i] += sample * sqrt(pan);      

      if (!(res = advanceReadIndex(fIndex, stepSize, state))) break;
    }
    break;
    
  case 2:
    
    for (i = 0; i < nFrames; i++) {

      if (!(res = vAdsr->tick(&s))) break;

      float sl,sr;

      getStereoFrame(&sl, &sr, (int)fIndex);
      
      sl *= velocity * s;
      sr *= velocity * s;

      l[i] += sl * sqrt(1.0f - pan);
      r[i] += sr * sqrt(pan);

      if (!(res = advanceReadIndex(fIndex, stepSize, state))) break;
    }
    
    break;
  }

  unlock();
  
  rIndex = (int)fIndex;

  return(res);
}


int Sample::getNoteOffset(int n) {

  return(n - midiNote);
}


void Sample::setMidiRangeStart(int note) {

  if (note <= midiNote) {
    
    midiRangeStart = note;
    instrument->sampleMap.clearSample(this);
    
    for (int i = midiRangeStart; i <= midiRangeEnd; i++) 
      instrument->sampleMap.setSample(i, this);
  }
  
  notify();
}

void Sample::setMidiRangeEnd(int note) {
  
  if (note >= midiNote) {
  
  midiRangeEnd = note;
  instrument->sampleMap.clearSample(this);

  for (int i = midiRangeStart; i <= midiRangeEnd; i++) 
    instrument->sampleMap.setSample(i, this);
  }

  notify();
}

int Sample::getMidiRangeStart() { return midiRangeStart; }
int Sample::getMidiRangeEnd() { return midiRangeEnd; }


/*
CVS stuff...

$Id: sample.cpp,v 1.23 2004-04-18 14:23:01 chenz Exp $ 
*/

