#ifndef _SAMPLE_H
#define _SAMPLE_H

#include <stdlib.h>
#include <stdio.h>

#include <jack/jack.h>

#define SAMPLE_DIVIDER (32767.0f)

#include "simsam.h"
#include "observer.h"
#include "envelope.h"

class Instrument;

class Sample : public Subject {

 private:

  float *W_data[2];

  int numFrames;
  int frameSize;
  int numChannels;

  static int maxId;
  static int getNewId();

  int loopStart;
  int loopEnd;
  int sampleStart;
  int sampleEnd;

  bool loop;
  bool middleLoop;
  bool trigger;
  bool exclusive;
  bool reverse;

  float gain;
  int iGain;

  float pan;
  int iPan;

  int midiNote;
  int midiRangeStart;
  int midiRangeEnd;

  Instrument *instrument;

  pthread_mutex_t mutex;

  int allocateBuffers();
  void freeBuffers();
  int deInterleave();

 public:

  EnvADSR adsr;

  int id;

  char path[1024];
  char name[256];

  int loadFile(const char *filename);
  void unload();

  float getMonoFrame(int pos);
  void getStereoFrame(float *l, float *r, int pos);

  /**
   * Advances the index i according to sample settings like loop points and
   * the given state
   */
  bool advanceReadIndex(int &i, int state);
  bool advanceReadIndex(float &fIndex, float stepSize, int state);

  /**
   * mixes nSamples frames of the sample, starting at rIndex, to the specified 
   * buffers. rIndex gets updated accordingly. Returns false if the sample has
   * ended.
   */
  bool addToMix(float *l, float *r, int nSamples, int &rIndex, int state, float velocity, EnvADSR *vAdsr, int noteOffset);

  Sample(Instrument *i);
  ~Sample();

  /**
   * used by the waveform viewer
   */
  void renderData(float **buffers, int numBuffers, int start, int end, int bufferLen);

  int getNumChannels();
  int getNumFrames();

  void setSampleStart(int v);
  void setSampleEnd(int v);
  void setLoopStart(int v);
  void setLoopEnd(int v);

  /**
   * sanity self-check
   */
  bool isValid();

  int getLoopStart();
  int getLoopEnd();
  int getSampleStart();
  int getSampleEnd();

  void setPan(int v);
  int getPan();

  void setGain(int v);
  int getGain();
  
  void setLoop(bool on);
  void setMiddleLoop(bool on);
  void setTrigger(bool on);
  void setExclusive(bool on);

  bool getLoop();
  bool getMiddleLoop();
  bool getTrigger();
  bool getExclusive();

  void setReverse(bool on);
  bool getReverse();

  void setMidiNote(int v);
  int getMidiNote();

  /**
   * Locks access to the sample's buffer using a mutex
   */
  void lock();

  /**
   * Unlocks access to the sample's buffer
   */
  void unlock();

  int getNoteOffset(int n);

  void setMidiRangeStart(int i);
  void setMidiRangeEnd(int i);

  int getMidiRangeStart();
  int getMidiRangeEnd();

  Instrument* getInstrument();
};

#endif // _SAMPLE_H


