#include "voice.h"
#include "error.h"

#include "simsam.h"
#include "output.h"

Voice::Voice() {

  state = OFF;
}

bool Voice::allocate(Sample *s, int oNum) {
  
  if (isUsed()) return(false);

  state = ALLOCATED;
  incVoiceCnt();
  
  sample = s;
  sample->adsr.copyTo(&adsr);

  output = Output::getOutput(oNum);

  //  error_message("voice allocated, output %d", oNum);

  return(true);
}

bool Voice::isUsed() {

  return(state != OFF);
}

void Voice::setState(int s) {

  if (!isUsed()) return;

  switch(s) {

  case OFF:

    state = OFF;
    decVoiceCnt();

    break;

  case ON:
    
    state = ON;
    adsr.noteOn();

    if (sample->getReverse()) readIndex = sample->getSampleEnd();
    else readIndex = sample->getSampleStart();

    break;

  case RELEASED:
    
    state = RELEASED;

    // release ADSR if not in 'loop while pressed' mode
    if (!(sample->getTrigger() && sample->getMiddleLoop())) adsr.noteOff();

    break;
  }
}


void Voice::addToMix(float *l, float *r, int len) {

  if(!isUsed()) return;

  bool sampleNotDone = sample->addToMix(l, r, len, readIndex, state, velocity, &adsr, noteOffset);
  if (!sampleNotDone) setState(OFF);
}


bool Voice::isExclusive() {

  if (isUsed()) return(sample->getExclusive());
  return(false);
}


void Voice::noteOn(int ch, int n, int vel) {

  int i;

  if (!isUsed()) { error_message("using unallocated sample!"); return; }

  if (state == ON) {

    if (sample->getTrigger() && sample->getLoop() && !sample->getMiddleLoop()) {
    
      // end 'loop until pressed'

      setState(RELEASED);
      return;
    }
  }

  velocity = ((float)vel) / 127.0f;
  noteOffset = sample->getNoteOffset(n);
  note = n;
  channel = ch;

  if (sample->getExclusive()) {

    // cancel other exclusive voices
    makeExclusive(channel, this);
  }

  setState(ON);
}


void Voice::noteOff() {

  if (!isUsed()) return; // voice already finished by itself

  if (!sample->getTrigger()) setState(RELEASED);
  else if (sample->getMiddleLoop()) setState(RELEASED);
}


int Voice::getNote() {

  if (!isUsed()) return(-1);
  return(note);
}


int Voice::getChannel() {
  
  if (!isUsed()) return(-1);
  return(channel);
}

Output* Voice::getOutput() {

  return(output);
}


// static stuff:

Voice Voice::voices[SIMSAM_MAX_VOICES];

Voice* Voice::allocateVoice(Sample *s, int oNum) {
  
  for (int i = 0; i < SIMSAM_MAX_VOICES; i++) {
    
    if (voices[i].allocate(s, oNum)) return(&voices[i]);
  }
  
  error_message("voice overflow!");
  return(NULL);
}

void Voice::makeExclusive(int channel, Voice *v) {

  for (int i = 0; i < SIMSAM_MAX_VOICES; i++) {

    if (voices[i].isExclusive()
	&& (voices[i].getChannel() == channel)
	&& (&voices[i] != v)) voices[i].setState(Voice::OFF);
  }
}


void Voice::panic() {
  
  for (int i = 0; i < SIMSAM_MAX_VOICES; i++) {

    voices[i].setState(Voice::OFF);
  }
}


Voice *Voice::getVoice(int channel, int note) {

  for (int i = 0; i < SIMSAM_MAX_VOICES; i++) {
    
    if ((voices[i].getNote() == note)
	&& (voices[i].getChannel() == channel)) return(&voices[i]);
  }

  return NULL;
}


int Voice::voiceCnt = 0;

void Voice::incVoiceCnt() {

  voiceCnt++;
  //  error_message("[%d] active voices", voiceCnt);
}

void Voice::decVoiceCnt() {

  voiceCnt--;
  //  error_message("[%d] active voices", voiceCnt);
}




/*
  $Id: voice.cpp,v 1.9 2008-02-20 10:06:23 chenz Exp $
*/
